"use strict";
// Copyright 2020-2025 SubQuery Pte Ltd authors & contributors
// SPDX-License-Identifier: GPL-3.0
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getAllJsonObjects = getAllJsonObjects;
exports.getAllEnums = getAllEnums;
exports.getAllEntitiesRelations = getAllEntitiesRelations;
exports.setJsonObjectType = setJsonObjectType;
const assert_1 = __importDefault(require("assert"));
const graphql_1 = require("graphql");
const array_1 = require("../array");
const logger_1 = require("../logger");
const types_1 = require("../types");
const constant_1 = require("./constant");
const schema_1 = require("./schema");
const types_2 = require("./types");
const logger = new logger_1.Logger({ level: 'info', outputFormat: 'colored' }).getLogger('Utils.entities');
function getAllJsonObjects(_schema) {
    return Object.values(getSchema(_schema).getTypeMap())
        .filter((node) => { var _a, _b; return (_b = (_a = node.astNode) === null || _a === void 0 ? void 0 : _a.directives) === null || _b === void 0 ? void 0 : _b.find(({ name: { value } }) => value === constant_1.DirectiveName.JsonField); })
        .map((node) => node)
        .filter(graphql_1.isObjectType);
}
function getAllEnums(_schema) {
    return getEnumsFromSchema(getSchema(_schema));
}
function getDirectives(schema, names) {
    const res = [];
    for (const name of names) {
        const directive = schema.getDirective(name);
        (0, assert_1.default)(directive, `${name} directive is required`);
        res.push(directive);
    }
    return res;
}
// eslint-disable-next-line complexity
function getAllEntitiesRelations(_schema) {
    var _a, _b, _c, _d, _e, _f, _g;
    if (_schema === null) {
        return {
            models: [],
            relations: [],
            enums: [],
        };
    }
    const schema = getSchema(_schema);
    const entities = Object.values(schema.getTypeMap())
        .filter((node) => { var _a, _b; return (_b = (_a = node.astNode) === null || _a === void 0 ? void 0 : _a.directives) === null || _b === void 0 ? void 0 : _b.find(({ name: { value } }) => value === constant_1.DirectiveName.Entity); })
        .filter(graphql_1.isObjectType);
    const jsonObjects = getAllJsonObjects(schema);
    const entityNameSet = entities.map((entity) => entity.name);
    const enums = new Map(getEnumsFromSchema(schema).map((node) => [
        node.name,
        { name: node.name, description: node.description, values: node.getValues().map((v) => v.value) },
    ]));
    const modelRelations = { models: [], relations: [], enums: [...enums.values()] };
    const [derivedFrom, indexDirective, idDbType] = getDirectives(schema, ['derivedFrom', 'index', 'dbType']);
    for (const entity of entities) {
        const newModel = {
            name: entity.name,
            description: (_a = entity.description) !== null && _a !== void 0 ? _a : undefined,
            fields: [],
            indexes: [],
        };
        const entityFields = Object.values(entity.getFields());
        const idField = entityFields.find((field) => field.name === 'id');
        if (!idField) {
            throw new Error(`Entity "${entity.name}" is missing required id field.`);
        }
        else if (idField.type.toString() !== 'ID!') {
            throw new Error(`Entity "${entity.name}" type must be ID, received ${idField.type.toString()}`);
        }
        const fkNameSet = [];
        for (const field of entityFields) {
            const typeString = extractType(field.type);
            const derivedFromDirectValues = field.astNode ? (0, graphql_1.getDirectiveValues)(derivedFrom, field.astNode) : undefined;
            const indexDirectiveVal = field.astNode ? (0, graphql_1.getDirectiveValues)(indexDirective, field.astNode) : undefined;
            const dbTypeDirectiveVal = field.astNode ? (0, graphql_1.getDirectiveValues)(idDbType, field.astNode) : undefined;
            //If is a basic scalar type
            const typeClass = (0, types_1.getTypeByScalarName)(typeString);
            if (typeClass === null || typeClass === void 0 ? void 0 : typeClass.fieldScalar) {
                newModel.fields.push(packEntityField(typeString, field, false));
            }
            // If is an enum
            else if (enums.has(typeString)) {
                newModel.fields.push({
                    type: typeString,
                    description: (_b = field.description) !== null && _b !== void 0 ? _b : undefined,
                    isEnum: true,
                    isArray: (0, graphql_1.isListType)((0, graphql_1.isNonNullType)(field.type) ? (0, graphql_1.getNullableType)(field.type) : field.type),
                    nullable: !(0, graphql_1.isNonNullType)(field.type),
                    name: field.name,
                });
            }
            // If is a foreign key
            else if (entityNameSet.includes(typeString) && !derivedFromDirectValues) {
                let relatedTypeString = types_2.FieldScalar.String;
                // Check to see if the relation ID type is modified with the @dbType directive
                const relatedIdField = (_c = entities.find((e) => e.name === typeString)) === null || _c === void 0 ? void 0 : _c.getFields().id;
                const dbIdType = (relatedIdField === null || relatedIdField === void 0 ? void 0 : relatedIdField.astNode) ? (0, graphql_1.getDirectiveValues)(idDbType, relatedIdField.astNode) : undefined;
                if (dbIdType) {
                    relatedTypeString = dbIdType.type;
                }
                newModel.fields.push(packEntityField(relatedTypeString, field, true));
                modelRelations.relations.push({
                    from: entity.name,
                    type: 'belongsTo',
                    to: typeString,
                    foreignKey: `${field.name}Id`,
                });
                // Ensures that foreign keys are linked correctly
                if (field.astNode &&
                    (field.astNode.type.kind === 'ListType' ||
                        (field.astNode.type.kind === 'NonNullType' && field.astNode.type.type.kind === 'ListType'))) {
                    const resolveName = (type) => {
                        switch (type.kind) {
                            case 'NamedType':
                                return type.name.value;
                            case 'NonNullType':
                                return resolveName(type.type);
                            case 'ListType':
                                return resolveName(type.type);
                            default:
                                // Any case in case future adds new kind
                                throw new Error(`Unhandled node kind: ${type.kind}`);
                        }
                    };
                    throw new Error(`Field "${field.name}" on entity "${newModel.name}" is missing "derivedFrom" directive. Please also make sure "${resolveName(field.astNode.type)}" has a field of type "${newModel.name}".`);
                }
                newModel.indexes.push({
                    unique: false,
                    fields: [`${field.name}Id`],
                    using: types_2.IndexType.HASH,
                });
                fkNameSet.push(field.name);
            }
            // If is derivedFrom
            else if (entityNameSet.includes(typeString) && derivedFromDirectValues) {
                modelRelations.relations.push({
                    from: entity.name,
                    type: (0, graphql_1.isListType)((0, graphql_1.isNonNullType)(field.type) ? (0, graphql_1.getNullableType)(field.type) : field.type) ? 'hasMany' : 'hasOne',
                    to: typeString,
                    foreignKey: `${derivedFromDirectValues.field}Id`,
                    fieldName: field.name,
                });
            }
            // If is jsonField
            else if (jsonObjects.map((json) => json.name).includes(typeString)) {
                const jsonObject = jsonObjects.find((object) => object.name === typeString);
                (0, assert_1.default)(jsonObject, `Json object ${typeString} not found`);
                const jsonObjectType = setJsonObjectType(jsonObject, jsonObjects);
                newModel.fields.push(packJSONField(typeString, field, jsonObjectType));
                const directive = (_e = (_d = jsonObject.astNode) === null || _d === void 0 ? void 0 : _d.directives) === null || _e === void 0 ? void 0 : _e.find(({ name: { value } }) => value === constant_1.DirectiveName.JsonField);
                const argValue = (_g = (_f = directive === null || directive === void 0 ? void 0 : directive.arguments) === null || _f === void 0 ? void 0 : _f.find((arg) => arg.name.value === 'indexed')) === null || _g === void 0 ? void 0 : _g.value;
                // For backwards compatibility if the argument is not defined then the index will be added
                if (!argValue || (isBooleanValueNode(argValue) && argValue.value !== false)) {
                    newModel.indexes.push({
                        unique: false,
                        fields: [field.name],
                        using: types_2.IndexType.GIN,
                    });
                }
            }
            else {
                throw new Error(`${typeString} is not a valid type`);
            }
            // handle indexes
            if (indexDirectiveVal) {
                if (typeString !== 'ID' && typeClass) {
                    newModel.indexes.push({
                        unique: indexDirectiveVal.unique,
                        fields: [field.name],
                    });
                }
                else if (typeString !== 'ID' && entityNameSet.includes(typeString)) {
                    if (indexDirectiveVal.unique) {
                        const fkIndex = newModel.indexes.find((idx) => idx.fields.length === 1 && idx.fields[0] === `${field.name}Id`);
                        if (fkIndex) {
                            fkIndex.unique = true;
                        }
                    }
                }
                else if (typeString !== 'ID' && enums.has(typeString)) {
                    newModel.indexes.push({
                        fields: [field.name],
                    });
                }
                else {
                    throw new Error(`index can not be added on field ${field.name}`);
                }
            }
            // Update id type if directive specified
            if (dbTypeDirectiveVal) {
                if (typeString !== 'ID') {
                    throw new Error(`dbType directive can only be added on 'id' field, received: ${field.name}`);
                }
                const dbType = dbTypeDirectiveVal.type;
                const t = (0, types_1.getTypeByScalarName)(dbType);
                // Allowlist of types that can be used.
                const allowedTypes = ['BigInt', 'Float', 'ID', 'Int', 'String'];
                if (!t || !allowedTypes.includes(t.name)) {
                    throw new Error(`${dbType} is not a defined scalar type, please use another type in the dbType directive.\nAvailable types: ${allowedTypes.join(', ')}`);
                }
                const f = newModel.fields.find((f) => f.name === 'id');
                if (!f) {
                    throw new Error('Expected id field to exist on model');
                }
                f.type = t.name;
            }
        }
        // Composite Indexes
        const compositeIndexDirective = schema.getDirective('compositeIndexes');
        (0, assert_1.default)(compositeIndexDirective, 'compositeIndexes directive is required');
        (0, assert_1.default)(entity.astNode, 'Entity astNode is required');
        const compositeIndexDirectiveVal = (0, graphql_1.getDirectiveValues)(compositeIndexDirective, entity.astNode);
        if ((compositeIndexDirectiveVal === null || compositeIndexDirectiveVal === void 0 ? void 0 : compositeIndexDirectiveVal.fields) && (compositeIndexDirectiveVal === null || compositeIndexDirectiveVal === void 0 ? void 0 : compositeIndexDirectiveVal.fields.length)) {
            const duplicateIndexes = (0, array_1.findDuplicateStringArray)(compositeIndexDirectiveVal.fields);
            if (duplicateIndexes.length) {
                throw new Error(`Found duplicate composite indexes ${JSON.stringify(duplicateIndexes)} on entity ${entity.name}`);
            }
            compositeIndexDirectiveVal.fields.forEach((indexFields) => {
                const joinFields = getJoinIndexFields(entity, entityFields, fkNameSet, indexFields);
                newModel.indexes.push({
                    fields: joinFields,
                });
            });
        }
        // Fulltext Search
        const fullTextDirective = schema.getDirective('fullText');
        (0, assert_1.default)(fullTextDirective, 'fullText directive is required');
        const fullTextDirectiveVal = (0, graphql_1.getDirectiveValues)(fullTextDirective, entity.astNode);
        if (fullTextDirectiveVal) {
            if (!fullTextDirectiveVal.fields.length) {
                throw new Error(`Expected fullText directive to have at least one field on entity ${entity.name}`);
            }
            // Make fields unique
            fullTextDirectiveVal.fields = [...new Set(fullTextDirectiveVal.fields)];
            fullTextDirectiveVal.fields.forEach((searchField, index) => {
                const field = newModel.fields.find((f) => [searchField, `${searchField}Id`].includes(f.name));
                if (!field) {
                    throw new Error(`Field "${searchField}" in fullText directive doesn't exist on entity "${entity.name}"`);
                }
                if (!['String', 'ID'].includes(field.type)) {
                    throw new Error(`fullText directive fields only supports String types`);
                }
                // If the field is a realation, we rename the field to include _id
                if (field.name === `${searchField}Id`) {
                    fullTextDirectiveVal.fields[index] = `${searchField}_id`;
                }
            });
            newModel.fullText = fullTextDirectiveVal;
        }
        modelRelations.models.push(newModel);
    }
    validateRelations(modelRelations);
    return modelRelations;
}
function getJoinIndexFields(entity, entityFields, fkNameSet, indexArgFields) {
    if (indexArgFields.length === 1) {
        logger.warn(`Composite index expected to be more than 1 field , entity ${entity} [${indexArgFields}]`);
    }
    if (indexArgFields.length > 3) {
        throw new Error(`Composite index on entity ${entity} expected not more than 3 fields, index [${indexArgFields}] got ${indexArgFields.length} fields`);
    }
    // check duplicate fields
    const duplicateFields = indexArgFields.filter((name, index, arr) => arr.indexOf(name) !== index);
    if (duplicateFields.length) {
        throw new Error(`Composite index ${entity}.${indexArgFields} got duplicated fields: ${duplicateFields}`);
    }
    return indexArgFields.map((j) => {
        const fieldInEntity = entityFields.find((f) => f.name === j);
        // check whether these fields are exist in the entity
        if (fieldInEntity === undefined) {
            throw new Error(`Composite index [${indexArgFields}], field ${j} not found within entity ${entity} `);
        }
        // check is fk
        if (fkNameSet.includes(fieldInEntity.name)) {
            return `${j}Id`;
        }
        return j;
    });
}
function packEntityField(typeString, field, isForeignKey) {
    var _a;
    return {
        name: isForeignKey ? `${field.name}Id` : field.name,
        type: typeString,
        description: (_a = field.description) !== null && _a !== void 0 ? _a : undefined,
        isArray: (0, graphql_1.isListType)((0, graphql_1.isNonNullType)(field.type) ? (0, graphql_1.getNullableType)(field.type) : field.type),
        nullable: !(0, graphql_1.isNonNullType)(field.type),
        isEnum: false,
    };
}
function packJSONField(typeString, field, jsonObject) {
    var _a;
    return {
        name: field.name,
        type: 'Json',
        description: (_a = field.description) !== null && _a !== void 0 ? _a : undefined,
        jsonInterface: jsonObject,
        isArray: (0, graphql_1.isListType)((0, graphql_1.isNonNullType)(field.type) ? (0, graphql_1.getNullableType)(field.type) : field.type),
        nullable: !(0, graphql_1.isNonNullType)(field.type),
        isEnum: false,
    };
}
function setJsonObjectType(jsonObject, jsonObjects) {
    const graphQLJsonObject = {
        name: jsonObject.name,
        fields: [],
    };
    for (const field of Object.values(jsonObject.getFields())) {
        //check if field is also json
        const typeString = extractType(field.type);
        const jsonType = jsonObjects.find((json) => json.name === typeString);
        graphQLJsonObject.fields.push({
            name: field.name,
            type: jsonType ? 'Json' : extractType(field.type),
            jsonInterface: jsonType ? setJsonObjectType(jsonType, jsonObjects) : undefined,
            nullable: !(0, graphql_1.isNonNullType)(field.type),
            isArray: (0, graphql_1.isListType)((0, graphql_1.isNonNullType)(field.type) ? (0, graphql_1.getNullableType)(field.type) : field.type),
        });
    }
    return graphQLJsonObject;
}
function getSchema(_schema) {
    return typeof _schema === 'string' ? (0, schema_1.buildSchemaFromFile)(_schema) : _schema;
}
function getEnumsFromSchema(schema) {
    return Object.values(schema.getTypeMap())
        .filter((r) => r.astNode !== undefined)
        .filter(graphql_1.isEnumType);
}
//Get the type, ready to be convert to string
function extractType(type) {
    if ((0, graphql_1.isUnionType)(type)) {
        throw new Error(`Not support Union type`);
    }
    if ((0, graphql_1.isInterfaceType)(type)) {
        throw new Error(`Not support Interface type`);
    }
    const offNullType = (0, graphql_1.isNonNullType)(type) ? (0, graphql_1.getNullableType)(type) : type;
    const offListType = (0, graphql_1.isListType)(offNullType) ? (0, graphql_1.assertListType)(offNullType).ofType : type;
    return (0, graphql_1.isNonNullType)(offListType)
        ? (0, graphql_1.getNullableType)(offListType).name
        : offListType.name;
}
function validateRelations(modelRelations) {
    for (const r of modelRelations.relations.filter((model) => model.type === 'hasMany' || model.type === 'hasOne')) {
        (0, assert_1.default)(modelRelations.models.find((model) => model.name === r.to && model.fields.find((field) => field.name === r.foreignKey)), `Please check entity ${r.from} with field ${r.fieldName} has correct relation with entity ${r.to}`);
    }
}
function isBooleanValueNode(valueNode) {
    return (valueNode === null || valueNode === void 0 ? void 0 : valueNode.kind) === 'BooleanValue';
}
//# sourceMappingURL=entities.js.map