"use strict";
// Copyright 2020-2025 SubQuery Pte Ltd authors & contributors
// SPDX-License-Identifier: GPL-3.0
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Logger = void 0;
const path_1 = __importDefault(require("path"));
const worker_threads_1 = require("worker_threads");
const flatted_1 = require("flatted");
const pino_1 = __importDefault(require("pino"));
const rotating_file_stream_1 = require("rotating-file-stream");
const colors_1 = require("./colors");
const constants_1 = require("./constants");
function formatErrorString(err, stack = false) {
    var _a;
    if (err instanceof Error) {
        const t = (_a = err.constructor.name) !== null && _a !== void 0 ? _a : 'Error';
        let formattedError = `${colors_1.ctx.red(`${t}:`)} ${colors_1.ctx.yellow(err.message)}`;
        if (stack) {
            formattedError += `\n${colors_1.ctx.red('Stack:')} ${colors_1.ctx.gray(err.stack)}`;
        }
        if (err.cause) {
            formattedError += `\n${colors_1.ctx.red('Cause:')} ${formatErrorString(err.cause, stack)}`;
        }
        return formattedError;
    }
    return String(err);
}
function formatErrorJson(err) {
    if (err instanceof Error) {
        return {
            type: 'error',
            name: err.name,
            message: err.message,
            stack: err.stack,
            cause: err.cause ? formatErrorJson(err.cause) : undefined,
        };
    }
    else {
        return (0, flatted_1.stringify)(err);
    }
}
class Logger {
    constructor({ filepath, level: logLevel = 'info', nestedKey, outputFormat, rotate, debugFilter = [] }) {
        this.childLoggers = {};
        this._debugFilter = debugFilter;
        const options = {
            messageKey: 'message',
            timestamp: () => `,"timestamp":"${new Date().toISOString()}"`,
            nestedKey,
            level: logLevel,
            formatters: {
                level(label, number) {
                    return { level: label };
                },
            },
            serializers: outputFormat === 'json'
                ? {
                    payload: formatErrorJson,
                }
                : {},
            prettyPrint: outputFormat !== 'json',
            prettifier: function (options) {
                // `this` is bound to the pino instance
                // Deal with whatever options are supplied.
                return function prettifier(inputData) {
                    let logObject;
                    if (typeof inputData === 'string') {
                        logObject = JSON.parse(inputData);
                    }
                    else if (isObject(inputData)) {
                        logObject = inputData;
                    }
                    if (!logObject)
                        return inputData;
                    // implement prettification
                    const { category, level, message, payload, time } = logObject;
                    let error = '';
                    if (payload instanceof Error) {
                        if (['debug', 'trace'].includes(logLevel)) {
                            error = `\n${formatErrorString(payload, true)}`;
                        }
                        else {
                            error = formatErrorString(payload);
                        }
                    }
                    return `${time} <${colors_1.ctx.magentaBright(category)}> ${(0, colors_1.colorizeLevel)(level)} ${message} ${error}\n`;
                };
                function isObject(input) {
                    return Object.prototype.toString.apply(input) === '[object Object]';
                }
            },
        };
        if (filepath) {
            const baseName = path_1.default.basename(filepath);
            const dirName = path_1.default.dirname(path_1.default.resolve(filepath));
            const rotateOptions = {
                interval: '1d',
                maxFiles: 7,
                maxSize: '1G',
            };
            if (rotate) {
                this.pino = (0, pino_1.default)(options, (0, rotating_file_stream_1.createStream)(baseName, Object.assign({ path: dirName }, rotateOptions)));
            }
            else {
                this.pino = (0, pino_1.default)(options, (0, rotating_file_stream_1.createStream)(baseName, { path: dirName }));
            }
        }
        else {
            this.pino = (0, pino_1.default)(options);
        }
    }
    getLogger(category) {
        if (!this.childLoggers[category]) {
            this.childLoggers[category] = this.pino.child({ category });
            this.applyChildDebug(category);
        }
        return this.childLoggers[category];
    }
    setLevel(level) {
        this.pino.level = level;
        Object.keys(this.childLoggers).map((key) => this.applyChildDebug(key));
    }
    setDebugFilter(debugFilter) {
        this._debugFilter = debugFilter;
        Object.keys(this.childLoggers).map((key) => this.applyChildDebug(key));
    }
    get debugFilter() {
        return this._debugFilter.map((f) => f.trim());
    }
    applyChildDebug(category) {
        if (!this.childLoggers[category].level)
            return;
        const checkCategory = worker_threads_1.isMainThread ? category : category.replace(`-#${worker_threads_1.threadId}`, '');
        if (this.debugFilter.includes(`-${checkCategory}`)) {
            this.pino.info(`Debug logging is disabled for ${category}`);
            // Set the log level to the global log level or INFO if its debug
            const newLevel = Math.max(constants_1.LEVELS_MAP[this.pino.level], constants_1.LEVELS_MAP.info);
            this.childLoggers[category].level = constants_1.LEVELS[newLevel].toLowerCase();
        }
        else if (this.debugFilter.includes(checkCategory)) {
            this.pino.info(`Debug logging is enabled for ${category}`);
            this.childLoggers[category].level = 'debug';
        }
        else if (this.debugFilter.includes('*')) {
            // Don't log wildcards, it spams the output
            this.childLoggers[category].level = 'debug';
        }
    }
}
exports.Logger = Logger;
//# sourceMappingURL=logger.js.map